#include <stddef.h>
#include <stdio.h>

#include <thread>
#include <chrono>
#include <fstream>
#include <sstream>
#include <cstring>
#include <cmath>

#include "TRPCLatWrapper.hpp"
#include "TRPCLAT_GeneralWrapper.h"
#include "dpc_trpc_lat/executable.h"
#ifdef __cplusplus
extern "C"
{
#endif

// C header here
#include "TRPC_Lat_CG.h"

#ifdef __cplusplus
}
#endif

typedef enum {
 IPC_Target_Motor_En,       // uint8        --> AccVlcMainReq = 1 
 IPC_Target_Gear,           // uint8        --> AccVlcMainReq = 1 -> D, else -> N
 IPC_Target_Mode,           // uint8        --> Set as 1
 IPC_Life_Signal,           // uint8        --> Counter 0-15
 IPC_Target_Speed,          // float32      --> Vx_des_tar
 IPC_Target_SpdValid,       // uint8        --> Set as 1
 IPC_Target_Brk_Press,      // double       --> Ax_tar < 0 and VlcDCEnable == 1
 IPC_Target_Angle,          // float32      --> EPSRoadWheelReqValue
 IPC_Target_Steering_En,    // uint8        --> EPSRoadWheelReqSts
 IPC_Target_Torque,         // double       --> MProp
 IPC_Torque_Rpm_Mode,       // uint8        --> Set 1  
 IPC_TurnLight,             // uint8        --> EPSRoadWheelReqValue < 0 -> Left, > 0 -> Right, brake prussre>50 -> 3, else ->0
 IPC_Target_SpdMode         // uint8        --> based on Vx_des_tar value range
} PendingSignals;

struct TRPCLatSWC2CANStruct
{
    SWC2CAN TargetSteeringAngle;
    boolean_T SteeringRequestModel;

    SWC2CAN SteeringRequest;
    float TargetSteeringAngleModel;

    SWC2CAN TurnLight;
    uint8_t TurnLightModel;
};


std::string LogFileNameDated(const std::string& inputName);
void logTRPCLatData(std::ofstream *ptr_File, uint8_t flag_First,
                    dpc::VED4DPC* const ptr_trpclat_data_from_VED,
                    CAN2SWC* const ptr_CAN2Lat_data,
                    dpc::Decision4Lat* const ptr_trpclat_data_from_DM,
                    TRPCLatSWC2CANStruct* const ptr_SWC2CAN_Lat_Data_log);
uint64_t getCurrentTimeStamp();

void assignTRPCLatInputsFromBA(dpc::Decision4Lat data_FromDM4TRPCLat) {
    TRPC_Lat_CG_U.BA_AlcDrvMd = 1; // Check with Roy data_FromDM4TRPCLat.ba_drive_md;
    TRPC_Lat_CG_U.BA_m_EgoLneWdth_sg = data_FromDM4TRPCLat.ba_ego_lane_width;
    TRPC_Lat_CG_U.BA_OptActionTyp_eBAAT = data_FromDM4TRPCLat.ba_opt_action;
    TRPC_Lat_CG_U.BA_OptCorridorID_sg = data_FromDM4TRPCLat.ba_opt_corridor_id;
    TRPC_Lat_CG_U.BA_RefLneC0_sg = data_FromDM4TRPCLat.ba_ref_line_c0;
    TRPC_Lat_CG_U.BA_RefLneC1_sg = data_FromDM4TRPCLat.ba_ref_line_c1;
    TRPC_Lat_CG_U.BA_RefLneC2_sg = data_FromDM4TRPCLat.ba_ref_line_c2;
    TRPC_Lat_CG_U.BA_RefLneC3_sg = data_FromDM4TRPCLat.ba_ref_line_c3;
    TRPC_Lat_CG_U.BA_RollingCounter = data_FromDM4TRPCLat.ba_rc4trp_lat; // Check with Roy
    TRPC_Lat_CG_U.BA_TRPSts_eTPRS = data_FromDM4TRPCLat.ba_trp_lat_sts; // Check with Roy



    std::cout << "@@@@@@ LAT IN: TRPC_Lat_CG_U.BA_AlcDrvMd: "<< TRPC_Lat_CG_U.BA_AlcDrvMd<< "\n";
    std::cout << "@@@@@@ LAT IN: TRPC_Lat_CG_U.BA_m_EgoLneWdth_sg: "<< TRPC_Lat_CG_U.BA_m_EgoLneWdth_sg<< "\n";
    std::cout << "@@@@@@ LAT IN: TRPC_Lat_CG_U.BA_OptActionTyp_eBAAT: "<< TRPC_Lat_CG_U.BA_OptActionTyp_eBAAT<< "\n";
    std::cout << "@@@@@@ LAT IN: TRPC_Lat_CG_U.BA_OptCorridorID_sg: "<< TRPC_Lat_CG_U.BA_OptCorridorID_sg<< "\n";
    std::cout << "@@@@@@ LAT IN: TRPC_Lat_CG_U.BA_RefLneC0_sg : "<< TRPC_Lat_CG_U.BA_RefLneC0_sg << "\n";
    std::cout << "@@@@@@ LAT IN: TRPC_Lat_CG_U.BA_RefLneC1_sg : "<< TRPC_Lat_CG_U.BA_RefLneC1_sg << "\n";
    std::cout << "@@@@@@ LAT IN: TRPC_Lat_CG_U.BA_RefLneC2_sg : "<< TRPC_Lat_CG_U.BA_RefLneC2_sg << "\n";
    std::cout << "@@@@@@ LAT IN: TRPC_Lat_CG_U.BA_RefLneC3_sg : "<< TRPC_Lat_CG_U.BA_RefLneC3_sg << "\n";
    std::cout << "@@@@@@ LAT IN: TRPC_Lat_CG_U.BA_RollingCounter: "<< TRPC_Lat_CG_U.BA_RollingCounter<< "\n";
    std::cout << "@@@@@@ LAT IN: TRPC_Lat_CG_U.BA_TRPSts_eTPRS: "<< TRPC_Lat_CG_U.BA_TRPSts_eTPRS<< "\n";
}

void assignTRPCLatInputsFromVED(const dpc::VED4DPC data_FromVED4TRPCLat) {
    TRPC_Lat_CG_U.VehSpdAvgDrn_Mps = data_FromVED4TRPCLat.LongitudinalVelocity;
    TRPC_Lat_CG_U.VSE_a_LatAcc_sg = data_FromVED4TRPCLat.LateralAcceleration;
    TRPC_Lat_CG_U.VSE_a_LgtAcclrt_sg = data_FromVED4TRPCLat.LongitudinalAcceleration;
    TRPC_Lat_CG_U.Vse_yr_YawRate_sg = data_FromVED4TRPCLat.Yawrate;
    std::cout << "@@@@@@ LAT IN: TRPC_Lat_CG_U.VehSpdAvgDrn_Mps: "<< TRPC_Lat_CG_U.VehSpdAvgDrn_Mps<< "\n";
    std::cout << "@@@@@@ LAT IN: TRPC_Lat_CG_U.VSE_a_LatAcc_sg: "<< TRPC_Lat_CG_U.VSE_a_LatAcc_sg<< "\n";
    std::cout << "@@@@@@ LAT IN: TRPC_Lat_CG_U.VSE_a_LgtAcclrt_sg: "<< TRPC_Lat_CG_U.VSE_a_LgtAcclrt_sg<< "\n";
    std::cout << "@@@@@@ LAT IN: TRPC_Lat_CG_U.Vse_yr_YawRate_sg: "<< TRPC_Lat_CG_U.Vse_yr_YawRate_sg<< "\n";
}

void assignTRPCLatInputsFromCAN(const CAN2SWC data_FromCAN4TRPCLat) {
    TRPC_Lat_CG_U.VCP_angD_RoadWhlAng_sg = -1 * (real_T)data_FromCAN4TRPCLat.VCU_Info_1_Signals.current_angle/57.3/14.0;//0.1
        // TO DO --> Does current_angle need any mapping?
    std::cout << "@@@@@@ LAT IN: TRPC_Lat_CG_U.VCP_angD_RoadWhlAng_sg: "<< TRPC_Lat_CG_U.VCP_angD_RoadWhlAng_sg<< "\n";
}


uint8_T TRPC_Lat_Run_Counter = 0;

const uint64_t TRPCLat_LogFileDuration = 240000; // milliseconds



int main() {
    dpc::VED4DPC trpclat_data_from_VED;
    dpc::Decision4Lat trpclat_data_from_DM;
    CAN2SWC CAN_Data;

    PendingSignals TargetSteeringAngle_enum = IPC_Target_Angle;
    SWC2CAN TargetSteeringAngle;

    PendingSignals SteeringRequest_enum = IPC_Target_Steering_En;
    SWC2CAN SteeringRequest;

    PendingSignals TurnLight_enum = IPC_TurnLight;
    SWC2CAN TurnLight;
    
    TRPCLatSWC2CANStruct LatToCANstruct4log;

    std::ofstream TRPCLatLogFile;
    TRPCLatLogFile.open(LogFileNameDated("TRPCLatLogFile"), std::ios::out);
    uint8_t flag_First = 1;

    uint32_t TRPCLat_run_counter = 0;

    uint64_t TRPCLat_dataLoggingTimerStart = getCurrentTimeStamp();

    ara::log::InitLogging("DPC", "EM_TEST", ara::log::LogLevel::kVerbose, (ara::log::LogMode::kConsole | ara::log::LogMode::kRemote));

	ara::log::Logger &mainLog { ara::log::CreateLogger("main", "em sample context", ara::log::LogLevel::kVerbose) };

	mainLog.LogInfo() << "Let's produce some trpc lat data...";

	GeneralWrapper gw(mainLog);

	std::cout << "OOOOOOOOOOOOOOOOOOOOOO LINE: " << __LINE__ <<std::endl;
	while (1)
    {
    	std::cout << "OOOOOOOOOOOOOOOOOOOOOO LINE: " << __LINE__ <<std::endl;
        if (TRPCLat_run_counter == 0) {
            TRPC_Lat_CG_initialize();
        }
        else {
            gw._VED2TRPCLat_Required.s.GetData(trpclat_data_from_VED);
            gw._DM2TRPCLat_Required.s.GetData(trpclat_data_from_DM);
            gw._CAN2Lat_Required.s.GetData(CAN_Data);
            // TO DO: Control data sync --> Must have the same arival time
            assignTRPCLatInputsFromVED(trpclat_data_from_VED);
            assignTRPCLatInputsFromBA(trpclat_data_from_DM);
            assignTRPCLatInputsFromCAN(CAN_Data);

            TRPC_Lat_CG_step();            

            
            //SteeringRequest.value = (UInt64)(TRPC_Lat_CG_Y.EPSRoadWheelReqSts);

            
            //TargetSteeringAngle.value = (UInt64)(TRPC_Lat_CG_Y.EPSRoadWheelReqValue); 
        }

        SteeringRequest.signal = (UInt8)(SteeringRequest_enum);
        // memcpy(&(SteeringRequest.value), &(TRPC_Lat_CG_Y.EPSRoadWheelReqSts), 
        //                             sizeof(SteeringRequest.value));
        SteeringRequest.value = static_cast<uint64_t>(TRPC_Lat_CG_Y.EPSRoadWheelReqSts);
        gw._TRPCLat2CAN_Provide.s.SetData(SteeringRequest);
        gw._TRPCLat2CAN_Provide.Publish();
        LatToCANstruct4log.SteeringRequest = SteeringRequest;
        LatToCANstruct4log.SteeringRequestModel = TRPC_Lat_CG_Y.EPSRoadWheelReqSts;
        std::cout << "SteeringRequest.value: "<< +SteeringRequest.value<<std::endl;

        TargetSteeringAngle.signal = (UInt8)(TargetSteeringAngle_enum);
        float temp_steering_angle = TRPC_Lat_CG_Y.EPSRoadWheelReqValue; //*(180./M_PI);
        double temp_steering_angle_double = -1*static_cast<double>(temp_steering_angle) * 57.3 * 14.0;
        memcpy(&(TargetSteeringAngle.value), &(temp_steering_angle_double), 
                                    sizeof(TargetSteeringAngle.value));
        gw._TRPCLat2CAN_Provide.s.SetData(TargetSteeringAngle);
        gw._TRPCLat2CAN_Provide.Publish();
        LatToCANstruct4log.TargetSteeringAngle = TargetSteeringAngle;
        LatToCANstruct4log.TargetSteeringAngleModel = temp_steering_angle;
        std::cout << "TargetSteeringAngle.value: "<< TargetSteeringAngle.value<<std::endl;
        std::cout << "TargetSteeringAngle main: "<< temp_steering_angle_double<<std::endl;
        std::cout << "temp_steering_angle "<< temp_steering_angle<<std::endl;
        std::cout << "TRPC_Lat_CG_Y.EPSRoadWheelReqValue "<< TRPC_Lat_CG_Y.EPSRoadWheelReqValue<<std::endl;

        TurnLight.signal = (UInt8)(TurnLight_enum);
        uint8_t turn_signal = 0;
        if (TRPC_Lat_CG_Y.EPSRoadWheelReqValue<0) {turn_signal=0;} // left or right steering is accomplished without turn light blinking
        else if (TRPC_Lat_CG_Y.EPSRoadWheelReqValue>0) {turn_signal=0;}
        // memcpy(&(TurnLight.value), &(turn_signal), 
        //                     sizeof(TurnLight.value));
        TurnLight.value = static_cast<uint64_t>(turn_signal);
        gw._TRPCLat2CAN_Provide.s.SetData(TurnLight);
        gw._TRPCLat2CAN_Provide.Publish();
        LatToCANstruct4log.TurnLight = TurnLight;
        LatToCANstruct4log.TurnLightModel = turn_signal;
        std::cout << "TurnLight.value: "<< +TurnLight.value<<std::endl;


        TRPCLat_run_counter++;

        uint64_t TRPCLat_dataLoggingTimerNow = getCurrentTimeStamp();
        if ((TRPCLat_dataLoggingTimerNow - TRPCLat_dataLoggingTimerStart)<= TRPCLat_LogFileDuration) {
            logTRPCLatData(&TRPCLatLogFile, flag_First, &trpclat_data_from_VED, &CAN_Data, 
                                    &trpclat_data_from_DM, &LatToCANstruct4log);
        }
        else {
            TRPCLatLogFile.close();
        }

        flag_First = 0;
        std::this_thread::sleep_for(std::chrono::milliseconds(10)); // TRPCLat cycle time: 10 ms
    }
    

    return 0;
}


std::string LogFileNameDated(const std::string& inputName) {
    time_t now = time(0);
    tm* localTime = localtime(&now);

    int year = 1900 + localTime->tm_year;
    int month = 1 + localTime->tm_mon;
    int day = localTime->tm_mday;
    int hour = localTime->tm_hour;
    int min = localTime->tm_min;
    int sec = localTime->tm_sec;

    std::string fileName = inputName + "_" + std::to_string(year) + "_" +
                           std::to_string(month) + "_" + std::to_string(day) + "_" +
                           std::to_string(hour) + "_" + std::to_string(min) + "_" +
                           std::to_string(sec) + ".csv";

    return fileName;
}
void logTRPCLatData(std::ofstream *ptr_File, uint8_t flag_First,
                    dpc::VED4DPC* const ptr_trpclat_data_from_VED,
                    CAN2SWC* const ptr_CAN2Lat_data,
                    dpc::Decision4Lat* const ptr_trpclat_data_from_DM,
                    TRPCLatSWC2CANStruct* const ptr_SWC2CAN_Lat_Data_log) {
                   
                   if (flag_First == 1) {
                        *ptr_File << "Timestamp" << ",";
                        // Log received data from CAN used in TRPCLat
                        *ptr_File << "VCU_Info_1_Signals.current_angle" << ",";
                        
                        // Log main received data from VED
                        *ptr_File << "LongitudinalVelocity" << ",";
                        *ptr_File << "LateralVelocity" << ",";
                        *ptr_File << "LongitudinalAcceleration" << ",";
                        *ptr_File << "LateralAcceleration" << ",";
                        *ptr_File << "MotionState" << ",";
                        *ptr_File << "Yawrate" << ",";
                        *ptr_File << "SteeringAngle" << ",";
                        *ptr_File << "SteeringRate" << ",";

                        // Log main received data from DM
                        *ptr_File << "ba_drive_md" << ",";
                        *ptr_File << "ba_ego_lane_width" << ",";
                        *ptr_File << "ba_fusion_time_stamp" << ",";
                        *ptr_File << "ba_hd_feel_md" << ",";
                        *ptr_File << "ba_lat_actv_func_sts" << ",";
                        *ptr_File << "ba_lat_fd_in_md" << ",";
                        *ptr_File << "ba_lat_fd_out_md" << ",";
                        *ptr_File << "ba_opt_action" << ",";
                        *ptr_File << "ba_opt_corridor_id" << ",";
                        *ptr_File << "ba_rc4trp_lat" << ",";
                        *ptr_File << "ba_ref_line_c0" << ",";
                        *ptr_File << "ba_ref_line_c1" << ",";
                        *ptr_File << "ba_ref_line_c2" << ",";
                        *ptr_File << "ba_ref_line_c3" << ",";
                        *ptr_File << "ba_ref_line_rng_ed" << ",";
                        *ptr_File << "ba_ref_line_rng_sts" << ",";
                        *ptr_File << "ba_trp_lat_sts" << ",";

                        // Log the CAN command to the actuator from TRPCLat
                        *ptr_File << "SteeringRequest.signal" << ",";
                        *ptr_File << "SteeringRequest.value" << ",";
                        *ptr_File << "SteeringRequestModel" << ",";
                        
                        *ptr_File << "TargetSteeringAngle.signal" << ",";
                        *ptr_File << "TargetSteeringAngle.value" << ",";
                        *ptr_File << "TargetSteeringAngleModel" << ",";

                        *ptr_File << "TurnLight.signal" << ",";
                        *ptr_File << "TurnLight.value" << ",";
                        *ptr_File << "TurnLightModel";

                        // Go to the next line
                        *ptr_File << "\n";
                   }

                   else {
                        *ptr_File << +getCurrentTimeStamp()<< ",";

                        // Log received data from CAN used in TRPCLat
                        *ptr_File << +ptr_CAN2Lat_data->VCU_Info_1_Signals.current_angle << ",";
                        
                        // Log main received data from VED
                        *ptr_File << +ptr_trpclat_data_from_VED->LongitudinalVelocity << ",";
                        *ptr_File << +ptr_trpclat_data_from_VED->LateralVelocity << ",";
                        *ptr_File << +ptr_trpclat_data_from_VED->LongitudinalAcceleration << ",";
                        *ptr_File << +ptr_trpclat_data_from_VED->LateralAcceleration << ",";
                        *ptr_File << +ptr_trpclat_data_from_VED->MotionState << ",";
                        *ptr_File << +ptr_trpclat_data_from_VED->Yawrate << ",";
                        *ptr_File << +ptr_trpclat_data_from_VED->SteeringAngle << ",";
                        *ptr_File << +ptr_trpclat_data_from_VED->SteeringRate << ",";

                        // Log main received data from DM
                        *ptr_File << +ptr_trpclat_data_from_DM->ba_drive_md << ",";
                        *ptr_File << +ptr_trpclat_data_from_DM->ba_ego_lane_width << ",";
                        *ptr_File << +ptr_trpclat_data_from_DM->ba_fusion_time_stamp << ",";
                        *ptr_File << +ptr_trpclat_data_from_DM->ba_hd_feel_md << ",";
                        *ptr_File << +ptr_trpclat_data_from_DM->ba_lat_actv_func_sts << ",";
                        *ptr_File << +ptr_trpclat_data_from_DM->ba_lat_fd_in_md << ",";
                        *ptr_File << +ptr_trpclat_data_from_DM->ba_lat_fd_out_md << ",";
                        *ptr_File << +ptr_trpclat_data_from_DM->ba_opt_action << ",";
                        *ptr_File << +ptr_trpclat_data_from_DM->ba_opt_corridor_id << ",";
                        *ptr_File << +ptr_trpclat_data_from_DM->ba_rc4trp_lat << ",";
                        *ptr_File << +ptr_trpclat_data_from_DM->ba_ref_line_c0 << ",";
                        *ptr_File << +ptr_trpclat_data_from_DM->ba_ref_line_c1 << ",";
                        *ptr_File << +ptr_trpclat_data_from_DM->ba_ref_line_c2 << ",";
                        *ptr_File << +ptr_trpclat_data_from_DM->ba_ref_line_c3 << ",";
                        *ptr_File << +ptr_trpclat_data_from_DM->ba_ref_line_rng_ed << ",";
                        *ptr_File << +ptr_trpclat_data_from_DM->ba_ref_line_rng_sts << ",";
                        *ptr_File << +ptr_trpclat_data_from_DM->ba_trp_lat_sts << ",";

                        // Log the CAN command to the actuator from TRPCLat
                        *ptr_File << +ptr_SWC2CAN_Lat_Data_log->SteeringRequest.signal << ",";
                        *ptr_File << +ptr_SWC2CAN_Lat_Data_log->SteeringRequest.value << ",";
                        *ptr_File << +ptr_SWC2CAN_Lat_Data_log->SteeringRequestModel << ",";
                        
                        *ptr_File << +ptr_SWC2CAN_Lat_Data_log->TargetSteeringAngle.signal<< ",";
                        *ptr_File << +ptr_SWC2CAN_Lat_Data_log->TargetSteeringAngle.value<< ",";
                        *ptr_File << +ptr_SWC2CAN_Lat_Data_log->TargetSteeringAngleModel<< ",";

                        *ptr_File << +ptr_SWC2CAN_Lat_Data_log->TurnLight.signal<< ",";
                        *ptr_File << +ptr_SWC2CAN_Lat_Data_log->TurnLight.value<< ",";
                        *ptr_File << +ptr_SWC2CAN_Lat_Data_log->TurnLightModel<< ",";

                        // Go to the next line
                        *ptr_File << "\n";
                   }
                }

uint64_t getCurrentTimeStamp() {
//    auto current_time = std::chrono::system_clock::now();
//    auto timestamp = std::chrono::duration_cast<std::chrono::seconds>(current_time.time_since_epoch());
//    uint32_t stamp_sec = static_cast<uint32_t>(timestamp.count());
//    uint32_t stamp_nanosec = static_cast<uint32_t>((current_time.time_since_epoch() - timestamp).count());
//    double CurrentTimeStamp = stamp_sec + (stamp_nanosec * 1e-9);
//    return CurrentTimeStamp;
	std::chrono::high_resolution_clock m_clock;
	return std::chrono::duration_cast<std::chrono::milliseconds>(m_clock.now().time_since_epoch()).count();
    }
